# -*- coding: utf-8 -*-
'''
Database object cache.

The shove psuedo-URL used for database object caches is the format used by
SQLAlchemy:

<driver>://<username>:<password>@<host>:<port>/<database>

<driver> is the database engine. The engines currently supported SQLAlchemy are
sqlite, mysql, postgres, oracle, mssql, and firebird.
<username> is the database account user name
<password> is the database accound password
<host> is the database location
<port> is the database port
<database> is the name of the specific database

For more information on specific databases see:

http://www.sqlalchemy.org/docs/dbengine.myt#dbengine_supported
'''

import time
import random
from datetime import datetime
try:
    from sqlalchemy import (
        MetaData, Table, Column, String, Binary, DateTime, bindparam, select,
        update, insert, delete,
    )
    from shove import DbBase
except ImportError:
    raise ImportError('Requires SQLAlchemy >= 0.4')

__all__ = ['DbCache']


class DbCache(DbBase):

    '''database cache backend'''

    def __init__(self, engine, **kw):
        super(DbCache, self).__init__(engine, **kw)
        # Get table name
        tablename = kw.get('tablename', 'cache')
        # Bind metadata
        self._metadata = MetaData(engine)
        # Make cache table
        self._store = Table(tablename, self._metadata,
            Column('key', String(60), primary_key=True, nullable=False),
            Column('value', Binary, nullable=False),
            Column('expires', DateTime, nullable=False),
        )
        # Create cache table if it does not exist
        if not self._store.exists(): self._store.create()
        # Set maximum entries
        self._max_entries = kw.get('max_entries', 300)
        # Maximum number of entries to cull per call if cache is full
        self._maxcull = kw.get('maxcull', 10)
        # Set timeout
        self.timeout = kw.get('timeout', 300)

    def __getitem__(self, key):
        row = select(
             [self._store.c.value, self._store.c.expires],
            self._store.c.key == key
        ).execute().fetchone()
        if row is not None:
            # Remove if item expired
            if row.expires < datetime.now().replace(microsecond=0):
                del self[key]
                raise KeyError(key)
            return self.loads(str(row.value))
        raise KeyError(key)

    def __setitem__(self, key, value):
        timeout, value, cache = self.timeout, self.dumps(value), self._store
        # Cull if too many items
        if len(self) >= self._max_entries:
            self._cull()
        # Generate expiration time
        expires = datetime.fromtimestamp(
            time.time() + timeout
        ).replace(microsecond=0)
        # Update database if key already present
        if key in self:
            update(
                cache,
                cache.c.key == key,
                dict(value=value, expires=expires),
            ).execute()
        # Insert new key if key not present
        else:
            insert(cache, dict(key=key, value=value, expires=expires)).execute()

    def _cull(self):
        '''Remove items in cache to make more room.'''
        cache, maxcull = self._store, self._maxcull
        # Remove items that have timed out
        now = datetime.now().replace(microsecond=0)
        delete(cache, cache.c.expires < now).execute()
        # Remove any items over the maximum allowed number in the cache
        if len(self) >= self._max_entries:
            # Upper limit for key query
            ul = maxcull * 2
            # Get list of keys
            keys = [
                i[0] for i in select(
                    [cache.c.key], limit=ul
                ).execute().fetchall()
            ]
            # Get some keys at random
            delkeys = list(random.choice(keys) for i in xrange(maxcull))
            # Delete keys
            fkeys = tuple({'key':k} for k in delkeys)
            delete(cache, cache.c.key.in_(bindparam('key'))).execute(*fkeys)



# -*- coding: utf-8 -*-
'''
DBM Database Store.

shove's psuedo-URL for DBM stores follows the form:

dbm://<path>

Where <path> is a URL path to a DBM database. Alternatively, the native
pathname to a DBM database can be passed as the 'engine' parameter.
'''

import anydbm

from shove.store import SyncStore


class DbmStore(SyncStore):

    '''Class for variants of the DBM database.'''

    init = 'dbm://'

    def __init__(self, engine, **kw):
        super(DbmStore, self).__init__(engine, **kw)
        self._store = anydbm.open(self._engine, 'c')
        try:
            self.sync = self._store.sync
        except AttributeError:
            pass


__all__ = ['DbmStore']



# -*- coding: utf-8 -*-
'''
Database object store.

The shove psuedo-URL used for database object stores is the format used by
SQLAlchemy:

<driver>://<username>:<password>@<host>:<port>/<database>

<driver> is the database engine. The engines currently supported SQLAlchemy are
sqlite, mysql, postgres, oracle, mssql, and firebird.
<username> is the database account user name
<password> is the database accound password
<host> is the database location
<port> is the database port
<database> is the name of the specific database

For more information on specific databases see:

http://www.sqlalchemy.org/docs/dbengine.myt#dbengine_supported
'''

try:
    from sqlalchemy import MetaData, Table, Column, String, Binary, select
    from shove import BaseStore, DbBase
except ImportError:
    raise ImportError('Requires SQLAlchemy >= 0.4')


class DbStore(BaseStore, DbBase):

    '''Database cache backend.'''

    def __init__(self, engine, **kw):
        super(DbStore, self).__init__(engine, **kw)
        # Get tablename
        tablename = kw.get('tablename', 'store')
        # Bind metadata
        self._metadata = MetaData(engine)
        # Make store table
        self._store = Table(tablename, self._metadata,
            Column('key', String(256), primary_key=True, nullable=False),
            Column('value', Binary, nullable=False),
        )
        # Create store table if it does not exist
        if not self._store.exists():
            self._store.create()

    def __getitem__(self, key):
        row = select(
            [self._store.c.value], self._store.c.key == key,
        ).execute().fetchone()
        if row is not None: return self.loads(str(row.value))
        raise KeyError(key)

    def __setitem__(self, k, v):
        v, store = self.dumps(v), self._store
        # Update database if key already present
        if k in self:
            store.update(store.c.key==k).execute(value=v)
        # Insert new key if key not present
        else:
            store.insert().execute(key=k, value=v)

    def keys(self):
        '''Returns a list of keys in the store.'''
        return list(i[0] for i in select(
            [self._store.c.key]
        ).execute().fetchall())


__all__ = ['DbStore']



# -*- coding: utf-8 -*-
'''
LevelDB Database Store.

shove's psuedo-URL for LevelDB stores follows the form:

leveldb://<path>

Where <path> is a URL path to a LevelDB database. Alternatively, the native
pathname to a LevelDB database can be passed as the 'engine' parameter.
'''

try:
    import leveldb
except ImportError:
    raise ImportError('This store requires py-leveldb library')

from shove.store import SyncStore


class LevelDBStore(SyncStore):

    '''LevelDB based store'''

    init = 'leveldb://'

    def __init__(self, engine, **kw):
        super(LevelDBStore, self).__init__(engine, **kw)
        self._store = leveldb.LevelDB(self._engine)

    def __getitem__(self, key):
        item = self.loads(self._store.Get(key))
        if item is not None:
            return item
        raise KeyError(key)

    def __setitem__(self, key, value):
        self._store.Put(key, self.dumps(value))

    def __delitem__(self, key):
        self._store.Delete(key)

    def keys(self):
        return list(k for k, _ in self._store.RangeIter())


__all__ = ['LevelDBStore']



# -*- coding: utf-8 -*-
'''
File-based cache

shove's psuedo-URL for file caches follows the form:

file://<path>

Where the path is a URL path to a directory on a local filesystem.
Alternatively, a native pathname to the directory can be passed as the 'engine'
argument.
'''

import time

from shove import FileBase
from shove.cache.simple import SimpleCache


class FileCache(FileBase, SimpleCache):

    '''File-based cache backend'''

    def __init__(self, engine, **kw):
        super(FileCache, self).__init__(engine, **kw)

    def __getitem__(self, key):
        try:
            exp, value = super(FileCache, self).__getitem__(key)
            # Remove item if time has expired.
            if exp < time.time():
                del self[key]
                raise KeyError(key)
            return value
        except:
            raise KeyError(key)

    def __setitem__(self, key, value):
        if len(self) >= self._max_entries: self._cull()
        super(FileCache, self).__setitem__(
            key, (time.time() + self.timeout, value)
        )


__all__ = ['FileCache']



# -*- coding: utf-8 -*-
'''
Single-process in-memory store.

The shove psuedo-URL for a simple store is:

simple://
'''

from shove import BaseStore, SimpleBase


class SimpleStore(SimpleBase, BaseStore):

    '''Single-process in-memory store.'''

    def __init__(self, engine, **kw):
        super(SimpleStore, self).__init__(engine, **kw)


__all__ = ['SimpleStore']



# -*- coding: utf-8 -*-
'''
Durus object database frontend.

shove's psuedo-URL for Durus stores follows the form:

durus://<path>


Where the path is a URL path to a durus FileStorage database. Alternatively, a
native pathname to a durus database can be passed as the 'engine' parameter.
'''

try:
    from durus.connection import Connection
    from durus.file_storage import FileStorage
except ImportError:
    raise ImportError('Requires Durus library')

from shove.store import SyncStore


class DurusStore(SyncStore):

    '''Class for Durus object database frontend.'''

    init = 'durus://'

    def __init__(self, engine, **kw):
        super(DurusStore, self).__init__(engine, **kw)
        self._db = FileStorage(self._engine)
        self._connection = Connection(self._db)
        self.sync = self._connection.commit
        self._store = self._connection.get_root()

    def close(self):
        '''Closes all open storage and connections.'''
        self.sync()
        self._db.close()
        super(DurusStore, self).close()


__all__ = ['DurusStore']



# -*- coding: utf-8 -*-
'''
Redis-based object store

The shove psuedo-URL for a redis-based store is:

redis://<host>:<port>/<db>
'''

import urlparse

try:
    import redis
except ImportError:
    raise ImportError('This store requires the redis library')

from shove.store import SyncStore


class RedisStore(SyncStore):

    '''Redis based store'''

    init = 'redis://'

    def __init__(self, engine, **kw):
        super(RedisStore, self).__init__(engine, **kw)
        spliturl = urlparse.urlsplit(engine)
        host, port = spliturl[1].split(':')
        db = spliturl[2].replace('/', '')
        self._store = redis.Redis(host, int(port), db)

    def __getitem__(self, key):
        item = super(RedisStore, self).__getitem__(key)
        if item is not None:
            return item
        raise KeyError(key)


__all__ = ['RedisStore']



# -*- coding: utf-8 -*-
'''
Single-process in-memory cache.

The shove psuedo-URL for a simple cache is:

simple://
'''

import time
import random

from shove import SimpleBase


class SimpleCache(SimpleBase):

    '''Single-process in-memory cache.'''

    def __init__(self, engine, **kw):
        super(SimpleCache, self).__init__(engine, **kw)
        # Get random seed
        random.seed()
        # Set maximum number of items to cull if over max
        self._maxcull = kw.get('maxcull', 10)
        # Set max entries
        self._max_entries = kw.get('max_entries', 300)
        # Set timeout
        self.timeout = kw.get('timeout', 300)

    def __getitem__(self, key):
        exp, value = super(SimpleCache, self).__getitem__(key)
        # Delete if item timed out.
        if exp < time.time():
            super(SimpleCache, self).__delitem__(key)
            raise KeyError(key)
        return value

    def __setitem__(self, key, value):
        # Cull values if over max # of entries
        if len(self) >= self._max_entries:
            self._cull()
        # Set expiration time and value
        exp = time.time() + self.timeout
        super(SimpleCache, self).__setitem__(key, (exp, value))

    def _cull(self):
        '''Remove items in cache to make room.'''
        num, maxcull = 0, self._maxcull
        # Cull number of items allowed (set by self._maxcull)
        for key in self.keys():
            # Remove only maximum # of items allowed by maxcull
            if num <= maxcull:
                # Remove items if expired
                try:
                    self[key]
                except KeyError:
                    num += 1
            else: break
        # Remove any additional items up to max # of items allowed by maxcull
        while len(self) >= self._max_entries and num <= maxcull:
            # Cull remainder of allowed quota at random
            del self[random.choice(self.keys())]
            num += 1


__all__ = ['SimpleCache']



# -*- coding: utf-8 -*-
'''
Zope Object Database store frontend.

shove's psuedo-URL for ZODB stores follows the form:

zodb:<path>


Where the path is a URL path to a ZODB FileStorage database. Alternatively, a
native pathname to a ZODB database can be passed as the 'engine' argument.
'''

try:
    import transaction
    from ZODB import FileStorage, DB
except ImportError:
    raise ImportError('Requires ZODB library')

from shove.store import SyncStore


class ZodbStore(SyncStore):

    '''ZODB store front end.'''

    init = 'zodb://'

    def __init__(self, engine, **kw):
        super(ZodbStore, self).__init__(engine, **kw)
        # Handle psuedo-URL
        self._storage = FileStorage.FileStorage(self._engine)
        self._db = DB(self._storage)
        self._connection = self._db.open()
        self._store = self._connection.root()
        # Keeps DB in synch through commits of transactions
        self.sync = transaction.commit

    def close(self):
        '''Closes all open storage and connections.'''
        self.sync()
        super(ZodbStore, self).close()
        self._connection.close()
        self._db.close()
        self._storage.close()


__all__ = ['ZodbStore']



# -*- coding: utf-8 -*-
'''
File-based LRU cache

shove's psuedo-URL for file caches follows the form:

file://<path>

Where the path is a URL path to a directory on a local filesystem.
Alternatively, a native pathname to the directory can be passed as the 'engine'
argument.
'''

from shove import FileBase
from shove.cache.simplelru import SimpleLRUCache


class FileCache(FileBase, SimpleLRUCache):

    '''File-based LRU cache backend'''


__all__ = ['FileCache']



# -*- coding: utf-8 -*-
'''Common object storage frontend.'''

import os
import zlib
import urllib
try:
    import cPickle as pickle
except ImportError:
    import pickle
from collections import deque

try:
    # Import store and cache entry points if setuptools installed
    import pkg_resources
    stores = dict((_store.name, _store) for _store in
        pkg_resources.iter_entry_points('shove.stores'))
    caches = dict((_cache.name, _cache) for _cache in
        pkg_resources.iter_entry_points('shove.caches'))
    # Pass if nothing loaded
    if not stores and not caches:
        raise ImportError()
except ImportError:
    # Static store backend registry
    stores = dict(
        bsddb='shove.store.bsdb:BsdStore',
        cassandra='shove.store.cassandra:CassandraStore',
        dbm='shove.store.dbm:DbmStore',
        durus='shove.store.durusdb:DurusStore',
        file='shove.store.file:FileStore',
        firebird='shove.store.db:DbStore',
        ftp='shove.store.ftp:FtpStore',
        leveldb='shove.store.leveldbstore:LevelDBStore',
        memory='shove.store.memory:MemoryStore',
        mssql='shove.store.db:DbStore',
        mysql='shove.store.db:DbStore',
        oracle='shove.store.db:DbStore',
        postgres='shove.store.db:DbStore',
        redis='shove.store.redisdb:RedisStore',
        s3='shove.store.s3:S3Store',
        simple='shove.store.simple:SimpleStore',
        sqlite='shove.store.db:DbStore',
        svn='shove.store.svn:SvnStore',
        zodb='shove.store.zodb:ZodbStore',
    )
    # Static cache backend registry
    caches = dict(
        bsddb='shove.cache.bsdb:BsdCache',
        file='shove.cache.file:FileCache',
        filelru='shove.cache.filelru:FileLRUCache',
        firebird='shove.cache.db:DbCache',
        memcache='shove.cache.memcached:MemCached',
        memlru='shove.cache.memlru:MemoryLRUCache',
        memory='shove.cache.memory:MemoryCache',
        mssql='shove.cache.db:DbCache',
        mysql='shove.cache.db:DbCache',
        oracle='shove.cache.db:DbCache',
        postgres='shove.cache.db:DbCache',
        simple='shove.cache.simple:SimpleCache',
        simplelru='shove.cache.simplelru:SimpleLRUCache',
        sqlite='shove.cache.db:DbCache',
    )


def getbackend(uri, engines, **kw):
    '''
    Loads the right backend based on a URI.

    @param uri Instance or name string
    @param engines A dictionary of scheme/class pairs
    @param kw Keywords'''
    if isinstance(uri, basestring):
        mod = engines[uri.split('://', 1)[0]]
        # Load module if setuptools not present
        if isinstance(mod, basestring):
            # Isolate classname from dot path
            module, klass = mod.split(':')
            # Load module
            mod = getattr(__import__(module, '', '', ['']), klass)
        # Load appropriate class from setuptools entry point
        else:
            mod = mod.load()
        # Return instance
        return mod(uri, **kw)
    # No-op for existing instances
    return uri


def synchronized(func):
    '''
    Decorator to lock and unlock a method (Phillip J. Eby).

    @param func Method to decorate
    '''
    def wrapper(self, *__args, **__kw):
        self._lock.acquire()
        try:
            return func(self, *__args, **__kw)
        finally:
            self._lock.release()
    wrapper.__name__ = func.__name__
    wrapper.__dict__ = func.__dict__
    wrapper.__doc__ = func.__doc__
    return wrapper


class Base(object):

    '''Base Mapping class.'''

    def __init__(self, engine, **kw):
        '''
        @keyword compress True, False, or an integer compression level (1-9).
        '''
        self._compress = kw.get('compress', False)
        self._protocol = kw.get('protocol', pickle.HIGHEST_PROTOCOL)

    def __getitem__(self, key):
        raise NotImplementedError()

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __delitem__(self, key):
        raise NotImplementedError()

    def __contains__(self, key):
        try:
            value = self[key]
        except KeyError:
            return False
        return True

    def get(self, key, default=None):
        '''
        Fetch a given key from the mapping. If the key does not exist,
        return the default.

        @param key Keyword of item in mapping.
        @param default Default value (default: None)
        '''
        try:
            return self[key]
        except KeyError:
            return default

    def dumps(self, value):
        '''Optionally serializes and compresses an object.'''
        # Serialize everything but ASCII strings
        value = pickle.dumps(value, protocol=self._protocol)
        if self._compress:
            level = 9 if self._compress is True else self._compress
            value = zlib.compress(value, level)
        return value

    def loads(self, value):
        '''Deserializes and optionally decompresses an object.'''
        if self._compress:
            try:
                value = zlib.decompress(value)
            except zlib.error:
                pass
        value = pickle.loads(value)
        return value


class BaseStore(Base):

    '''Base Store class (based on UserDict.DictMixin).'''

    def __init__(self, engine, **kw):
        super(BaseStore, self).__init__(engine, **kw)
        self._store = None

    def __cmp__(self, other):
        if other is None:
            return False
        if isinstance(other, BaseStore):
            return cmp(dict(self.iteritems()), dict(other.iteritems()))

    def __del__(self):
        # __init__ didn't succeed, so don't bother closing
        if not hasattr(self, '_store'):
            return
        self.close()

    def __iter__(self):
        for k in self.keys():
            yield k

    def __len__(self):
        return len(self.keys())

    def __repr__(self):
        return repr(dict(self.iteritems()))

    def close(self):
        '''Closes internal store and clears object references.'''
        try:
            self._store.close()
        except AttributeError:
            pass
        self._store = None

    def clear(self):
        '''Removes all keys and values from a store.'''
        for key in self.keys():
            del self[key]

    def items(self):
        '''Returns a list with all key/value pairs in the store.'''
        return list(self.iteritems())

    def iteritems(self):
        '''Lazily returns all key/value pairs in a store.'''
        for k in self:
            yield (k, self[k])

    def iterkeys(self):
        '''Lazy returns all keys in a store.'''
        return self.__iter__()

    def itervalues(self):
        '''Lazily returns all values in a store.'''
        for _, v in self.iteritems():
            yield v

    def keys(self):
        '''Returns a list with all keys in a store.'''
        raise NotImplementedError()

    def pop(self, key, *args):
        '''
        Removes and returns a value from a store.

        @param args Default to return if key not present.
        '''
        if len(args) > 1:
            raise TypeError('pop expected at most 2 arguments, got ' + repr(
                1 + len(args))
            )
        try:
            value = self[key]
        # Return default if key not in store
        except KeyError:
            if args:
                return args[0]
        del self[key]
        return value

    def popitem(self):
        '''Removes and returns a key, value pair from a store.'''
        try:
            k, v = self.iteritems().next()
        except StopIteration:
            raise KeyError('Store is empty.')
        del self[k]
        return (k, v)

    def setdefault(self, key, default=None):
        '''
        Returns the value corresponding to an existing key or sets the
        to key to the default and returns the default.

        @param default Default value (default: None)
        '''
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return default

    def update(self, other=None, **kw):
        '''
        Adds to or overwrites the values in this store with values from
        another store.

        other Another store
        kw Additional keys and values to store
        '''
        if other is None:
            pass
        elif hasattr(other, 'iteritems'):
            for k, v in other.iteritems():
                self[k] = v
        elif hasattr(other, 'keys'):
            for k in other.keys():
                self[k] = other[k]
        else:
            for k, v in other:
                self[k] = v
        if kw:
            self.update(kw)

    def values(self):
        '''Returns a list with all values in a store.'''
        return list(v for _, v in self.iteritems())


class Shove(BaseStore):

    '''Common object frontend class.'''

    def __init__(self, store='simple://', cache='simple://', **kw):
        super(Shove, self).__init__(store, **kw)
        # Load store
        self._store = getbackend(store, stores, **kw)
        # Load cache
        self._cache = getbackend(cache, caches, **kw)
        # Buffer for lazy writing and setting for syncing frequency
        self._buffer, self._sync = dict(), kw.get('sync', 2)

    def __getitem__(self, key):
        '''Gets a item from shove.'''
        try:
            return self._cache[key]
        except KeyError:
            # Synchronize cache and store
            self.sync()
            value = self._store[key]
            self._cache[key] = value
            return value

    def __setitem__(self, key, value):
        '''Sets an item in shove.'''
        self._cache[key] = self._buffer[key] = value
        # When the buffer reaches self._limit, writes the buffer to the store
        if len(self._buffer) >= self._sync:
            self.sync()

    def __delitem__(self, key):
        '''Deletes an item from shove.'''
        try:
            del self._cache[key]
        except KeyError:
            pass
        self.sync()
        del self._store[key]

    def keys(self):
        '''Returns a list of keys in shove.'''
        self.sync()
        return self._store.keys()

    def sync(self):
        '''Writes buffer to store.'''
        for k, v in self._buffer.iteritems():
            self._store[k] = v
        self._buffer.clear()

    def close(self):
        '''Finalizes and closes shove.'''
        # If close has been called, pass
        if self._store is not None:
            try:
                self.sync()
            except AttributeError:
                pass
            self._store.close()
            self._store = self._cache = self._buffer = None


class FileBase(Base):

    '''Base class for file based storage.'''

    def __init__(self, engine, **kw):
        super(FileBase, self).__init__(engine, **kw)
        if engine.startswith('file://'):
            engine = urllib.url2pathname(engine.split('://')[1])
        self._dir = engine
        # Create directory
        if not os.path.exists(self._dir):
            self._createdir()

    def __getitem__(self, key):
        # (per Larry Meyn)
        try:
            item = open(self._key_to_file(key), 'rb')
            data = item.read()
            item.close()
            return self.loads(data)
        except:
            raise KeyError(key)

    def __setitem__(self, key, value):
        # (per Larry Meyn)
        try:
            item = open(self._key_to_file(key), 'wb')
            item.write(self.dumps(value))
            item.close()
        except (IOError, OSError):
            raise KeyError(key)

    def __delitem__(self, key):
        try:
            os.remove(self._key_to_file(key))
        except (IOError, OSError):
            raise KeyError(key)

    def __contains__(self, key):
        return os.path.exists(self._key_to_file(key))

    def __len__(self):
        return len(os.listdir(self._dir))

    def _createdir(self):
        '''Creates the store directory.'''
        try:
            os.makedirs(self._dir)
        except OSError:
            raise EnvironmentError(
                'Cache directory "%s" does not exist and ' \
                'could not be created' % self._dir
            )

    def _key_to_file(self, key):
        '''Gives the filesystem path for a key.'''
        return os.path.join(self._dir, urllib.quote_plus(key))

    def keys(self):
        '''Returns a list of keys in the store.'''
        return [urllib.unquote_plus(name) for name in os.listdir(self._dir)]


class SimpleBase(Base):

    '''Single-process in-memory store base class.'''

    def __init__(self, engine, **kw):
        super(SimpleBase, self).__init__(engine, **kw)
        self._store = dict()

    def __getitem__(self, key):
        try:
            return self._store[key]
        except:
            raise KeyError(key)

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        try:
            del self._store[key]
        except:
            raise KeyError(key)

    def __len__(self):
        return len(self._store)

    def keys(self):
        '''Returns a list of keys in the store.'''
        return self._store.keys()


class LRUBase(SimpleBase):

    def __init__(self, engine, **kw):
        super(LRUBase, self).__init__(engine, **kw)
        self._max_entries = kw.get('max_entries', 300)
        self._hits = 0
        self._misses = 0
        self._queue = deque()
        self._refcount = dict()

    def __getitem__(self, key):
        try:
            value = super(LRUBase, self).__getitem__(key)
            self._hits += 1
        except KeyError:
            self._misses += 1
            raise
        self._housekeep(key)
        return value

    def __setitem__(self, key, value):
        super(LRUBase, self).__setitem__(key, value)
        self._housekeep(key)
        if len(self._store) > self._max_entries:
            while len(self._store) > self._max_entries:
                k = self._queue.popleft()
                self._refcount[k] -= 1
                if not self._refcount[k]:
                    super(LRUBase, self).__delitem__(k)
                    del self._refcount[k]

    def _housekeep(self, key):
        self._queue.append(key)
        self._refcount[key] = self._refcount.get(key, 0) + 1
        if len(self._queue) > self._max_entries * 4:
            self._purge_queue()

    def _purge_queue(self):
        for i in [None] * len(self._queue):
            k = self._queue.popleft()
            if self._refcount[k] == 1:
                self._queue.append(k)
            else:
                self._refcount[k] -= 1


class DbBase(Base):

    '''Database common base class.'''

    def __init__(self, engine, **kw):
        super(DbBase, self).__init__(engine, **kw)

    def __delitem__(self, key):
        self._store.delete(self._store.c.key == key).execute()

    def __len__(self):
        return self._store.count().execute().fetchone()[0]


__all__ = ['Shove']



# -*- coding: utf-8 -*-

import unittest

from shove import Shove


class TestRedisStore(unittest.TestCase):

    def setUp(self):
        self.store = Shove('redis://localhost:6379/0')

    def tearDown(self):
        self.store.clear()
        self.store.close()

    def test__getitem__(self):
        self.store['max'] = 3
        self.assertEqual(self.store['max'], 3)

    def test__setitem__(self):
        self.store['max'] = 3
        self.assertEqual(self.store['max'], 3)

    def test__delitem__(self):
        self.store['max'] = 3
        del self.store['max']
        self.assertEqual('max' in self.store, False)

    def test_get(self):
        self.store['max'] = 3
        self.assertEqual(self.store.get('min'), None)

    def test__cmp__(self):
        tstore = Shove()
        self.store['max'] = 3
        tstore['max'] = 3
        self.assertEqual(self.store, tstore)

    def test__len__(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.assertEqual(len(self.store), 2)

    def test_clear(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.clear()
        self.assertEqual(len(self.store), 0)

    def test_items(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        slist = list(self.store.items())
        self.assertEqual(('min', 6) in slist, True)

    def test_iteritems(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        slist = list(self.store.iteritems())
        self.assertEqual(('min', 6) in slist, True)

    def test_iterkeys(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        slist = list(self.store.iterkeys())
        self.assertEqual('min' in slist, True)

    def test_itervalues(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        slist = list(self.store.itervalues())
        self.assertEqual(6 in slist, True)

    def test_pop(self):
        self.store['max'] = 3
        self.store['min'] = 6
        item = self.store.pop('min')
        self.assertEqual(item, 6)

    def test_popitem(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        item = self.store.popitem()
        self.assertEqual(len(item) + len(self.store), 4)

    def test_setdefault(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['powl'] = 7
        self.store.setdefault('pow', 8)
        self.assertEqual(self.store.setdefault('pow', 8), 8)
        self.assertEqual(self.store['pow'], 8)

    def test_update(self):
        tstore = Shove()
        tstore['max'] = 3
        tstore['min'] = 6
        tstore['pow'] = 7
        self.store['max'] = 2
        self.store['min'] = 3
        self.store['pow'] = 7
        self.store.update(tstore)
        self.assertEqual(self.store['min'], 6)

    def test_values(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        slist = self.store.values()
        self.assertEqual(6 in slist, True)

    def test_keys(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        slist = self.store.keys()
        self.assertEqual('min' in slist, True)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''
subversion managed store.

The shove psuedo-URL used for a subversion store that is password protected is:

svn:<username><password>:<path>?url=<url>

or for non-password protected repositories:

svn://<path>?url=<url>

<path> is the local repository copy
<url> is the URL of the subversion repository
'''

import os
import urllib
import threading

try:
    import pysvn
except ImportError:
    raise ImportError('Requires Python Subversion library')

from shove import BaseStore, synchronized


class SvnStore(BaseStore):

    '''Class for subversion store.'''

    def __init__(self, engine=None, **kw):
        super(SvnStore, self).__init__(engine, **kw)
        # Get path, url from keywords if used
        path, url = kw.get('path'), kw.get('url')
        # Get username. password from keywords if used
        user, password = kw.get('user'), kw.get('password')
        # Process psuedo URL if used
        if engine is not None:
            path, query = engine.split('n://')[1].split('?')
            url = query.split('=')[1]
            # Check for username, password
            if '@' in path:
                auth, path = path.split('@')
                user, password = auth.split(':')
            path = urllib.url2pathname(path)
        # Create subversion client
        self._client = pysvn.Client()
        # Assign username, password
        if user is not None:
            self._client.set_username(user)
        if password is not None:
            self._client.set_password(password)
        # Verify that store exists in repository
        try:
            self._client.info2(url)
        # Create store in repository if it doesn't exist
        except pysvn.ClientError:
            self._client.mkdir(url, 'Adding directory')
        # Verify that local copy exists
        try:
            if self._client.info(path) is None:
                self._client.checkout(url, path)
        # Check it out if it doesn't exist
        except pysvn.ClientError:
            self._client.checkout(url, path)
        self._path, self._url = path, url
        # Lock
        self._lock = threading.Condition()

    @synchronized
    def __getitem__(self, key):
        try:
            return self.loads(self._client.cat(self._key_to_file(key)))
        except:
            raise KeyError(key)

    @synchronized
    def __setitem__(self, key, value):
        fname = self._key_to_file(key)
        # Write value to file
        open(fname, 'wb').write(self.dumps(value))
        # Add to repository
        if key not in self:
            self._client.add(fname)
        self._client.checkin([fname], 'Adding %s' % fname)

    @synchronized
    def __delitem__(self, key):
        try:
            fname = self._key_to_file(key)
            self._client.remove(fname)
            # Remove deleted value from repository
            self._client.checkin([fname], 'Removing %s' % fname)
        except:
            raise KeyError(key)

    def _key_to_file(self, key):
        '''Gives the filesystem path for a key.'''
        return os.path.join(self._path, urllib.quote_plus(key))

    @synchronized
    def keys(self):
        '''Returns a list of keys in the subversion repository.'''
        return list(str(i.name.split('/')[-1]) for i
            in self._client.ls(self._path))


__all__ = ['SvnStore']



# -*- coding: utf-8 -*-
'''
Thread-safe in-memory cache.

The shove psuedo-URL for a memory cache is:

memory://
'''

import copy
import threading

from shove import synchronized
from shove.cache.simple import SimpleCache


class MemoryCache(SimpleCache):

    '''Thread-safe in-memory cache backend.'''

    def __init__(self, engine, **kw):
        super(MemoryCache, self).__init__(engine, **kw)
        self._lock = threading.Condition()

    @synchronized
    def __setitem__(self, key, value):
        super(MemoryCache, self).__setitem__(key, value)

    @synchronized
    def __getitem__(self, key):
        return copy.deepcopy(super(MemoryCache, self).__getitem__(key))

    @synchronized
    def __delitem__(self, key):
        super(MemoryCache, self).__delitem__(key)


__all__ = ['MemoryCache']



# -*- coding: utf-8 -*-
'''
S3-accessed stores

shove's psuedo-URL for stores found on Amazon.com's S3 web service follows this
form:

s3://<s3_key>:<s3_secret>@<bucket>

<s3_key> is the Access Key issued by Amazon
<s3_secret> is the Secret Access Key issued by Amazon
<bucket> is the name of the bucket accessed through the S3 service
'''

try:
    from boto.s3.connection import S3Connection
    from boto.s3.key import Key
except ImportError:
    raise ImportError('Requires boto library')

from shove import BaseStore


class S3Store(BaseStore):

    def __init__(self, engine=None, **kw):
        super(S3Store, self).__init__(engine, **kw)
        # key = Access Key, secret=Secret Access Key, bucket=bucket name
        key, secret, bucket = kw.get('key'), kw.get('secret'), kw.get('bucket')
        if engine is not None:
            auth, bucket = engine.split('://')[1].split('@')
            key, secret = auth.split(':')
        # kw 'secure' = (True or False, use HTTPS)
        self._conn = S3Connection(key, secret, kw.get('secure', False))
        buckets = self._conn.get_all_buckets()
        # Use bucket if it exists
        for b in buckets:
            if b.name == bucket:
                self._store = b
                break
        # Create bucket if it doesn't exist
        else:
            self._store = self._conn.create_bucket(bucket)
        # Set bucket permission ('private', 'public-read',
        # 'public-read-write', 'authenticated-read'
        self._store.set_acl(kw.get('acl', 'private'))
        # Updated flag used for avoiding network calls
        self._updated, self._keys = True, None

    def __getitem__(self, key):
        rkey = self._store.lookup(key)
        if rkey is None:
            raise KeyError(key)
        # Fetch string
        value = self.loads(rkey.get_contents_as_string())
        # Flag that the store has not been updated
        self._updated = False
        return value

    def __setitem__(self, key, value):
        rkey = Key(self._store)
        rkey.key = key
        rkey.set_contents_from_string(self.dumps(value))
        # Flag that the store has been updated
        self._updated = True

    def __delitem__(self, key):
        try:
            self._store.delete_key(key)
            # Flag that the store has been updated
            self._updated = True
        except:
            raise KeyError(key)

    def keys(self):
        '''Returns a list of keys in the store.'''
        return list(i[0] for i in self.items())

    def items(self):
        '''Returns a list of items from the store.'''
        if self._updated or self._keys is None:
            self._keys = self._store.get_all_keys()
        return list((str(k.key), k) for k in self._keys)

    def iteritems(self):
        '''Lazily returns items from the store.'''
        for k in self.items():
            yield (k.key, k)


__all__ = ['S3Store']



# -*- coding: utf-8 -*-
'''
Filesystem-based object store

shove's psuedo-URL for filesystem-based stores follows the form:

file://<path>

Where the path is a URL path to a directory on a local filesystem.
Alternatively, a native pathname to the directory can be passed as the 'engine'
argument.
'''

from shove import BaseStore, FileBase


class FileStore(FileBase, BaseStore):

    '''File-based store.'''

    def __init__(self, engine, **kw):
        super(FileStore, self).__init__(engine, **kw)


__all__ = ['FileStore']



# -*- coding: utf-8 -*-

import os
import time
import unittest

from shove.cache.file import FileCache


class TestFileCache(unittest.TestCase):

    initstring = 'file://test'
    cacheclass = FileCache

    def setUp(self):
        self.cache = self.cacheclass(self.initstring)

    def tearDown(self):
        self.cache = None
        for x in os.listdir('test'): os.remove(os.path.join('test', x))
        os.rmdir('test')

    def test_getitem(self):
        self.cache['test'] = 'test'
        self.assertEqual(self.cache['test'], 'test')

    def test_setitem(self):
        self.cache['test'] = 'test'
        self.assertEqual(self.cache['test'], 'test')

    def test_delitem(self):
        self.cache['test'] = 'test'
        del self.cache['test']
        self.assertEqual('test' in self.cache, False)

    def test_get(self):
        self.assertEqual(self.cache.get('min'), None)

    def test_timeout(self):
        cache = self.cacheclass(self.initstring, timeout=1)
        cache['test'] = 'test'
        time.sleep(2)

        def tmp():
            cache['test']
        self.assertRaises(KeyError, tmp)

    def test_cull(self):
        cache = self.cacheclass(self.initstring, max_entries=1)
        cache['test'] = 'test'
        cache['test2'] = 'test'
        num = len(cache)
        self.assertEquals(num, 1)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''
Single-process in-memory LRU cache.

The shove psuedo-URL for a simple cache is:

simplelru://
'''

from shove import LRUBase


class SimpleLRUCache(LRUBase):

    '''In-memory cache that purges based on least recently used item.'''


__all__ = ['SimpleLRUCache']



# -*- coding: utf-8 -*-
'''setup - setuptools based setup for shove.'''

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='shove',
    version='0.4.0',
    description='''Common object storage frontend''',
    long_description=open('README.rst').read(),
    author='L. C. Rees',
    author_email='lcrees@gmail.com',
    url='https://bitbucket.org/lcrees/shove/',
    license='BSD',
    packages=['shove', 'shove.cache', 'shove.store', 'shove.tests'],
    py_modules=['ez_setup'],
    test_suite='shove.tests',
    zip_safe=False,
    keywords='object storage persistence database shelve',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Database :: Front-Ends',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    entry_points='''
    [shove.stores]
    bsddb=shove.store.bsdb:BsdStore
    cassandra=shove.store.cassandra:CassandraStore
    dbm=shove.store.dbm:DbmStore
    durus=shove.store.durusdb:DurusStore
    file=shove.store.file:FileStore
    firebird=shove.store.db:DbStore
    ftp=shove.store.ftp:FtpStore
    leveldb=shove.store.leveldbstore:LevelDBStore
    memory=shove.store.memory:MemoryStore
    mssql=shove.store.db:DbStore
    mysql=shove.store.db:DbStore
    oracle=shove.store.db:DbStore
    postgres=shove.store.db:DbStore
    redis=shove.store.redisdb:RedisStore
    s3=shove.store.s3:S3Store
    simple=shove.store.simple:SimpleStore
    sqlite=shove.store.db:DbStore
    svn=shove.store.svn:SvnStore
    zodb=shove.store.zodb:ZodbStore
    [shove.caches]
    bsddb=shove.cache.bsdb:BsdCache
    file=shove.cache.file:FileCache
    filelru=shove.cache.filelru:FileLRUCache
    firebird=shove.cache.db:DbCache
    memcache=shove.cache.memcached:MemCached
    memlru=shove.cache.memlru:MemoryLRUCache
    memory=shove.cache.memory:MemoryCache
    mssql=shove.cache.db:DbCache
    mysql=shove.cache.db:DbCache
    oracle=shove.cache.db:DbCache
    postgres=shove.cache.db:DbCache
    simple=shove.cache.simple:SimpleCache
    simplelru=shove.cache.simplelru:SimpleLRUCache
    sqlite=shove.cache.db:DbCache
    '''
)



# -*- coding: utf-8 -*-

from urllib import url2pathname
from shove.store.simple import SimpleStore


class SyncStore(SimpleStore):

    '''Base class for stores where updates have to be committed.'''

    def __init__(self, engine, **kw):
        super(SyncStore, self).__init__(engine, **kw)
        if engine.startswith(self.init):
            self._engine = url2pathname(engine.split('://')[1])

    def __getitem__(self, key):
        return self.loads(super(SyncStore, self).__getitem__(key))

    def __setitem__(self, key, value):
        super(SyncStore, self).__setitem__(key, self.dumps(value))
        try:
            self.sync()
        except AttributeError:
            pass

    def __delitem__(self, key):
        super(SyncStore, self).__delitem__(key)
        try:
            self.sync()
        except AttributeError:
            pass


__all__ = [
    'bsdb', 'db', 'dbm', 'durusdb', 'file', 'ftp', 'memory', 's3', 'simple',
    'svn', 'zodb', 'redisdb'
]



# -*- coding: utf-8 -*-
'''
Berkeley Source Database Store.

shove's psuedo-URL for BSDDB stores follows the form:

bsddb://<path>

Where the path is a URL path to a Berkeley database. Alternatively, the native
pathname to a Berkeley database can be passed as the 'engine' parameter.
'''
try:
    import bsddb
except ImportError:
    raise ImportError('requires bsddb library')

import threading

from shove import synchronized
from shove.store import SyncStore


class BsdStore(SyncStore):

    '''Class for Berkeley Source Database Store.'''

    init = 'bsddb://'

    def __init__(self, engine, **kw):
        super(BsdStore, self).__init__(engine, **kw)
        self._store = bsddb.hashopen(self._engine)
        self._lock = threading.Condition()
        self.sync = self._store.sync

    @synchronized
    def __getitem__(self, key):
        return super(BsdStore, self).__getitem__(key)

    @synchronized
    def __setitem__(self, key, value):
        super(BsdStore, self).__setitem__(key, value)

    @synchronized
    def __delitem__(self, key):
        super(BsdStore, self).__delitem__(key)


__all__ = ['BsdStore']



# -*- coding: utf-8 -*-
'''
Cassandra-based object store

The shove psuedo-URL for a cassandra-based store is:

cassandra://<host>:<port>/<keyspace>/<columnFamily>
'''

import urlparse

try:
    import pycassa
except ImportError:
    raise ImportError('This store requires the pycassa library')

from shove import BaseStore


class CassandraStore(BaseStore):

    '''Cassandra based store'''

    init = 'cassandra://'

    def __init__(self, engine, **kw):
        super(CassandraStore, self).__init__(engine, **kw)
        spliturl = urlparse.urlsplit(engine)
        _, keyspace, column_family = spliturl[2].split('/')
        try:
            self._pool = pycassa.connect(keyspace, [spliturl[1]])
            self._store = pycassa.ColumnFamily(self._pool, column_family)
        except pycassa.InvalidRequestException:
            from pycassa.system_manager import SystemManager
            system_manager = SystemManager(spliturl[1])
            system_manager.create_keyspace(
                keyspace, 
                pycassa.system_manager.SIMPLE_STRATEGY, 
                {'replication_factor': str(kw.get('replication', 1))}
            )
            system_manager.create_column_family(keyspace, column_family)
            self._pool = pycassa.connect(keyspace, [spliturl[1]])
            self._store = pycassa.ColumnFamily(self._pool, column_family)

    def __getitem__(self, key):
        try:
            item = self._store.get(key).get(key)
            if item is not None:
                return self.loads(item)
            raise KeyError(key)
        except pycassa.NotFoundException:
            raise KeyError(key)

    def __setitem__(self, key, value):
        self._store.insert(key, dict(key=self.dumps(value)))

    def __delitem__(self, key):
        # beware eventual consistency
        try:
            self._store.remove(key)
        except pycassa.NotFoundException:
            raise KeyError(key)

    def clear(self):
        # beware eventual consistency
        self._store.truncate()

    def keys(self):
        return list(i[0] for i in self._store.get_range())


__all__ = ['CassandraStore']
